/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

#ifndef HSS_PAR_EXTRACTION_HPP
#define HSS_PAR_EXTRACTION_HPP

template<typename T, typename S> void HSS_tree<T,S>::HSS_par_Extraction(T *B, int * descB, int *I, int nI, int *J, int nJ, gridinfo *ginfo) {
  /* Sequential element extraction: B(i,j) is an approximation to A(I(i),J(j)),
   * obtained from the HSS form of A.
   * B is assumed to be allocated.
   *
   * The algorithm is essentially a matrix-vector with sparse RHS and sparse
   * solution: A(i,j)=(A e_j)(i).
   *
   * We "prune" the tree by computing which elements need to be traversed:
   *  - For each element J(j) of J, the set of nodes that needs to be processed
   *  in the bottom-up traversal of the tree is the path from the unique leaf
   *  that contains J(j) up to the root node.
   *  - For each element I(i) of I, the set of nodes that needs to be processed
   *  in the top-traversal of the tree is the path from the unique leaf that
   *  contains I(i) up to the root node.
   *
   * In the bottom-up traversal, each node has a set of active columns:
   *  - For a leaf, it corresponds to elements of J that are in the index set
   *    of that leaf.
   *  - For non-leaves, it is the union of the sets of the children.
   * In the top-down traversal
   *
   * Algorithm:
   * Step1: postorder traversal:
   *     if i leaf
   *       scan J to compute the number of actives columns at i
   *       if non-zero,
   *         y{i}=V{i}'*[e_j1...e_jnJ]
   *     otherwise, y{i}=V{i}'*[y{child1}; y{child2}]
   *                where y{child1} or y{child2} (but not both) can be 0
   *                if child1 or child2 wasn't traversed.
   *                y{child1} and y{child2} have different column counts;
   *                the merge is actually
   *                [y{child1} 0; 0 y{child2}]
   * Step2: reverse postorder traversal
   *   scan J to compute isneeded: isneeded[i]=1 if i needs to be
   *   traversed
   *   if i needs to be traversed:
   *     if i leaf, B(I,:)=U{i}*z{i}+D{i}*x(I,:)
   *     otherwise,
   *       r1/2: row-rank for child1/2
   *       z{child1}=B12{i}*y{child2}+U{i}(1:r1)*z{i}
   *       z{child2}=B21{i}*y{child1}+U{i}(r1+1:r1+r2)*z{i}
   *       (if root node, the second term in the sum is not here).
   *
   * The parallel version follows the same flow as the parallel compression;
   * there are three kinds of nodes to distinguish: serial subtrees, leaves
   * that are mapped on several processes, and parallel interior nodes.
   *
   */

  int i, j, k;
  int mb, nb;
  int br, b;
  int rows_y, rows_z, rows_y1, rows_y2, rows_z1, rows_z2;
  int rows_B, cols_B;
  int istart, iend;
  T *Bl=NULL;
  T *tmp, *tmp2, *tmp3;
  T **y, **z;
  int *children1, *children2;
  int ch1, ch2;
  bool *active;
  int *columns;
  int myid, np;
  int proc_b, proc_g;
  int idummy, ierr;
  int ldl;
  int *descY=NULL, *descZ=NULL, descGen[BLACSCTXTSIZE], descBl[BLACSCTXTSIZE], *desc=NULL, descPad[BLACSCTXTSIZE];
  int locr, locc;
  HSS_node<T,S> *child1, *child2;
  gridinfo tmpinfo, tmpinfo1, tmpinfo2;

  mb=ginfo->mb;
  nb=ginfo->nb;
  MPI_Comm_rank(comm,&myid);
  MPI_Comm_size(comm,&np);

  /* Here y and z are not fields of a node but arrays indexed following the
   * postorder; the pointers "children" of a node are not enough to find the
   * right in y and z, what we need is numbers in the postorder. "children1"
   * and "children2" are the same information as children but they provide
   * numbers (indices in the tree).
   *
   * children1: children[i]==j: node number j is left child of node number i.
   * child1: pointer to left child (type HSS_node) of node nodes[i].
   */

  children1=new int[dim_tree];
  for(i=0;i<dim_tree;i++) {
    children1[i]=-1;
    if(nodes[i]->nr_children!=0)
      for(j=0;j<dim_tree;j++)
        if(nodes[j]==nodes[i]->children[0]) {
          children1[i]=j;
          break;
        }
  }
  children2=new int[dim_tree];
  for(i=0;i<dim_tree;i++) {
    children2[i]=-1;
    if(nodes[i]->nr_children!=0)
      for(j=0;j<dim_tree;j++)
        if(nodes[j]==nodes[i]->children[1]) {
          children2[i]=j;
          break;
        }
  }

  /* Step 1: postorder traversal (except root) */
  columns=new int[dim_tree];
  for(i=0;i<dim_tree;i++) {
    if(nodes[i]->nr_children==0) {
      columns[i]=0;
      for(j=0;j<nJ;j++)
        /* Is J(j) in the index set of i? */
        if(nodes[i]->tree_ib<=J[j] && J[j]<=nodes[i]->tree_ie)
          columns[i]++;
    } else {
      ch1=children1[i];
      ch2=children2[i];
      columns[i]=columns[ch1]+columns[ch2];
    }
  }

  y=new T*[dim_tree-1];
  for(i=0;i<dim_tree-1;i++)
    y[i]=NULL;

  /* Serial subtree */
  if(subtree[2*myid]>=0) {
    istart=subtree[2*myid];
    iend=std::min(subtree[2*myid+1],dim_tree-2);
    for(i=istart;i<=iend;i++) {
      /* Traverse subtree, excluding root node */
      b=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
      rows_y=nodes[i]->v_rank;
      if(nodes[i]->nr_children==0) {
        /* Leaf: y{i}=V{i}'*x(I,:) */
        if(!columns[i])
          continue; /* Node is not active */
        y[i]=new T[rows_y*columns[i]]();
        tmp=new T[b*columns[i]]();
        columns[i]=0;
        for(j=0;j<nJ;j++)
          if(nodes[i]->tree_ib<=J[j] && J[j]<=nodes[i]->tree_ie) {
            tmp[J[j]-nodes[i]->tree_ib+b*columns[i]]=ONE;
            columns[i]++;
          }
        nodes[i]->seq_applyGenerator('C','V',columns[i],ONE,tmp,b,ZERO,y[i],rows_y);
        delete[] tmp;
      } else {
        /* Non-leaf: y{i}=V{i}'*[y{child1} 0; 0 y{child2}] */
        if(!columns[i])
          continue; /* Node is not active */
        child1=nodes[i]->children[0];
        child2=nodes[i]->children[1];
        ch1=children1[i];
        ch2=children2[i];

        rows_y1=child1->v_rank;
        rows_y2=child2->v_rank;
        y[i]=new T[rows_y*columns[i]]();
        tmp=new T[(rows_y1+rows_y2)*columns[i]]();

        columns[i]=0;
        columns[ch1]=0;
        columns[ch2]=0;
        for(j=0;j<nJ;j++)
          if(nodes[i]->tree_ib<=J[j] && J[j]<=nodes[i]->tree_ie) {
            if(child1->tree_ib<=J[j] && J[j]<=child1->tree_ie) {
              /* From left child */
              lacpy('N',rows_y1,IONE,&y[ch1][rows_y1*columns[ch1]],rows_y1,&tmp[(rows_y1+rows_y2)*columns[i]],rows_y1+rows_y2);
              columns[i]++;
              columns[ch1]++;
            } else {
              /* From right child */
              lacpy('N',rows_y2,IONE,&y[ch2][rows_y2*columns[ch2]],rows_y2,&tmp[rows_y1+(rows_y1+rows_y2)*columns[i]],rows_y1+rows_y2);
              columns[i]++;
              columns[ch2]++;
            }
          }

        nodes[i]->seq_applyGenerator('C','V',columns[i],ONE,tmp,rows_y1+rows_y2,ZERO,y[i],rows_y);
        delete[] tmp;
      }
    }
    if(np>1) {
      rows_y=nodes[iend]->v_rank;
      idummy=std::max(1,numroc_(&rows_y,&mb,&nodes[iend]->myrow,&IZERO,&nodes[iend]->nprow));
      descY=new int[BLACSCTXTSIZE];
      descinit_(descY,&rows_y,&columns[iend],&mb,&nb,&IZERO,&IZERO,&(nodes[iend]->context),&idummy,&ierr);
    }
  }

  /* Leaves (avoid the case leaf=root) */
  for(i=0;i<dim_tree-1;i++) {
    if(   nodes[i]->procs[1]>nodes[i]->procs[0]
       && nodes[i]->nr_children==0
       && nodes[i]->procs[0]+nodes[i]->nprow*nodes[i]->npcol-1>=myid
       && nodes[i]->procs[0]<=myid
       && columns[i]
      ) {
      tmpinfo.ctxt=nodes[i]->context;
      tmpinfo.nprow=nodes[i]->nprow;
      tmpinfo.npcol=nodes[i]->npcol;
      tmpinfo.myrow=nodes[i]->myrow;
      tmpinfo.mycol=nodes[i]->mycol;
      tmpinfo.mb=ginfo->mb;
      tmpinfo.nb=ginfo->nb;

      b=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
      locr=numroc_(&b,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
      locc=numroc_(&columns[i],&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
      tmp=new T[locr*locc]();
      idummy=std::max(1,locr);
      desc=new int[BLACSCTXTSIZE];
      descinit_(desc,&b,&columns[i],&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
      columns[i]=0;
      for(j=0;j<nJ;j++)
        if(nodes[i]->tree_ib<=J[j] && J[j]<=nodes[i]->tree_ie) {
          columns[i]++;
          idummy=J[j]-nodes[i]->tree_ib+1;
          if(   nodes[i]->myrow==indxg2p_(&idummy,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow)
             && nodes[i]->mycol==indxg2p_(&columns[i],&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol)
          ) {
            idummy=J[j]-nodes[i]->tree_ib+1;
            tmp[(indxg2l_(&columns[i],&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol)-1)*locr+indxg2l_(&idummy,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow)-1]=ONE;
          }
        }

      rows_y=nodes[i]->v_rank;
      locr=numroc_(&rows_y,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
      locc=numroc_(&columns[i],&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
      y[i]=new T[locr*locc]();
      idummy=std::max(1,locr);
      descY=new int[BLACSCTXTSIZE];
      descinit_(descY,&rows_y,&columns[i],&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);

      nodes[i]->par_applyGenerator('C','V',columns[i],ONE,tmp,desc,ZERO,y[i],descY,&tmpinfo);
      delete[] tmp;
      delete[] desc;

    }
  }

  /* Interior nodes */
  for(j=0;j<par_lvl;j++) {
    i=par_nodes[j];
    if(!columns[i])
      continue;
    if(i==dim_tree-1)
      break;
    child1=nodes[i]->children[0];
    child2=nodes[i]->children[1];
    ch1=children1[i];
    ch2=children2[i];
    proc_b=nodes[i]->procs[0];
    proc_g=nodes[i]->procs[0]+nodes[i]->nprow*nodes[i]->npcol-1;
    tmpinfo.ctxt=nodes[i]->context;
    tmpinfo.nprow=nodes[i]->nprow;
    tmpinfo.npcol=nodes[i]->npcol;
    tmpinfo.myrow=nodes[i]->myrow;
    tmpinfo.mycol=nodes[i]->mycol;
    tmpinfo.mb=ginfo->mb;
    tmpinfo.nb=ginfo->nb;
    rows_y1=child1->v_rank;
    rows_y2=child2->v_rank;
    /* Merge y{child1} and y{child2} into tmp2.
     * Here the complication comes from the fact
     * that y{child1} and y{child2} have different
     * sizes and columns indices. We pad them
     * with zeros before merging them. This
     * allows us to reuse mergeVector.
     * If J was sorted we could maybe do something
     * better.
     */
    if(myid<child1->procs[0]+child1->nprow*child1->npcol) {
      /* Mapped on child 1; create tmp = y[ch1] with padded zeros */
      locr=numroc_(&rows_y1,&mb,&nodes[ch1]->myrow,&IZERO,&nodes[ch1]->nprow);
      locc=numroc_(&columns[i],&nb,&nodes[ch1]->mycol,&IZERO,&nodes[ch1]->npcol);
      tmp=new T[locr*locc]();
      idummy=std::max(1,locr);
      descinit_(descPad,&rows_y1,&columns[i],&mb,&nb,&IZERO,&IZERO,&nodes[ch1]->context,&idummy,&ierr);
      columns[i]=0;
      columns[ch1]=0;
      for(k=0;k<nJ;k++)
        if(nodes[i]->tree_ib<=J[k] && J[k]<=nodes[i]->tree_ie) {
          columns[i]++;
          if(child1->tree_ib<=J[k] && J[k]<=child1->tree_ie) {
            /* Copy columns corresponding to J[k]: columns[ch1] for child1,
             * columns[i] for tmp (same structure as parent i).
             */
            columns[ch1]++;
            pgeadd('N',rows_y1,IONE,ONE,y[ch1],IONE,columns[ch1],descY,ZERO,tmp,IONE,columns[i],descPad);
           }
        }
    } else if(myid>=child2->procs[0] && myid<child2->procs[0]+child2->nprow*child2->npcol) {
      /* Mapped on child 2; create tmp = y[ch2] with padded zeros */
      locr=numroc_(&rows_y2,&mb,&nodes[ch2]->myrow,&IZERO,&nodes[ch2]->nprow);
      locc=numroc_(&columns[i],&nb,&nodes[ch2]->mycol,&IZERO,&nodes[ch2]->npcol);
      tmp=new T[locr*locc]();
      idummy=std::max(1,locr);
      descinit_(descPad,&rows_y2,&columns[i],&mb,&nb,&IZERO,&IZERO,&nodes[ch2]->context,&idummy,&ierr);
      columns[i]=0;
      columns[ch2]=0;
      for(k=0;k<nJ;k++)
        if(nodes[i]->tree_ib<=J[k] && J[k]<=nodes[i]->tree_ie) {
          columns[i]++;
          if(child2->tree_ib<=J[k] && J[k]<=child2->tree_ie) {
            /* Copy columns corresponding to J[k]: columns[ch2] for child2,
             * columns[i] for tmp (same structure as parent i).
             */
            columns[ch2]++;
            pgeadd('N',rows_y2,IONE,ONE,y[ch2],IONE,columns[ch2],descY,ZERO,tmp,IONE,columns[i],descPad);
           }
        }
    } else
      /* Out of both grids */
      tmp=NULL;
    mergeVector(tmp,child1->procs[0],child1->procs[0]+child1->nprow*child1->npcol-1,rows_y1,IONE,child2->procs[0],child2->procs[0]+child2->nprow*child2->npcol-1,rows_y2,IONE,proc_b,proc_g,columns[i],descPad,&tmp2,&desc,nodes[i]->allctxt,&tmpinfo,comm);
    delete[] tmp;

    /* Exclude processes that are not in the grid of i */
    if(myid>=proc_b && myid<=proc_g) {
      rows_y=nodes[i]->v_rank;
      locr=numroc_(&rows_y,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
      locc=numroc_(&columns[i],&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
      y[i]=new T[locr*locc]();
      idummy=std::max(1,locr);
      if(descY==NULL)
        descY=new int[BLACSCTXTSIZE];
      descinit_(descY,&rows_y,&columns[i],&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
      nodes[i]->par_applyGenerator('C','V',columns[i],ONE,tmp2,desc,ZERO,y[i],descY,&tmpinfo);
      delete[] tmp2;
    }
  }

  /* Step 2: reverse postorder traversal */
  z=new T*[dim_tree-1];
  for(i=0;i<dim_tree-1;i++)
    z[i]=NULL;

  active=new bool[dim_tree];
  /* First, we determine active nodes using a bottom-up traversal */
  for(i=0;i<dim_tree;i++) {
    if(nodes[i]->nr_children==0) {
      active[i]=false;
      for(j=0;j<nI;j++)
        if(nodes[i]->tree_ib<=I[j] && I[j]<=nodes[i]->tree_ie) {
          active[i]=true;
          break;
        }
    } else {
      ch1=children1[i];
      ch2=children2[i];
      active[i]=active[ch1] || active[ch2];
    }
  }

  /* Parallel leaves (i.e. leaves mapped on multiple processes or roots
   * of serial subtrees) write their requested entries into Bl.
   * At the end, we merge all the Bl's to construct B.
   */
  if(subtree[2*myid]>=0) {
    /* Serial subtree */
    istart=subtree[2*myid];
    iend=subtree[2*myid+1];
    ldl=np==1?a_dim:nodes[iend]->tree_ie-nodes[iend]->tree_ib+1;
    Bl=new T[ldl*nJ]();
    descinit_(descBl,&ldl,&nJ,&mb,&nb,&IZERO,&IZERO,&nodes[iend]->context,&ldl,&ierr);
  }
  for(i=0;i<dim_tree;i++) {
    if(   nodes[i]->procs[1]>nodes[i]->procs[0]
       && nodes[i]->nr_children==0
       && nodes[i]->procs[0]+nodes[i]->nprow*nodes[i]->npcol-1>=myid
       && nodes[i]->procs[0]<=myid
       && active[i]
      ) {
      /* Active parallel leaf */
      tmpinfo.ctxt=nodes[i]->context;
      tmpinfo.nprow=nodes[i]->nprow;
      tmpinfo.npcol=nodes[i]->npcol;
      tmpinfo.myrow=nodes[i]->myrow;
      tmpinfo.mycol=nodes[i]->mycol;
      tmpinfo.mb=ginfo->mb;
      tmpinfo.nb=ginfo->nb;

      ldl=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
      locr=numroc_(&ldl,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
      locc=numroc_(&nJ,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
      Bl=new T[locr*locc]();
      idummy=std::max(1,locr);
      descinit_(descBl,&ldl,&nJ,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
    }
  }

  /* Interior nodes */
  for(j=par_lvl-1;j>=0;j--) {
    i=par_nodes[j];
    if(!active[i])
      continue; /* Node is not active */
    child1=nodes[i]->children[0];
    child2=nodes[i]->children[1];
    ch1=children1[i];
    ch2=children2[i];
    proc_b=nodes[i]->procs[0];
    proc_g=nodes[i]->procs[0]+nodes[i]->nprow*nodes[i]->npcol-1;
    tmpinfo.ctxt=nodes[i]->context;
    tmpinfo.nprow=nodes[i]->nprow;
    tmpinfo.npcol=nodes[i]->npcol;
    tmpinfo.myrow=nodes[i]->myrow;
    tmpinfo.mycol=nodes[i]->mycol;
    tmpinfo.mb=ginfo->mb;
    tmpinfo.nb=ginfo->nb;
    rows_z1=child1->u_rank;
    rows_z2=child2->u_rank;
    rows_y1=child1->v_rank;
    rows_y2=child2->v_rank;
    br=rows_z1+rows_z2;

    /* We need to compute [0 B12; B21 0]*[y1;y2]. B12 and B21 are
     * distributed on the processes mapped on i. The easy and
     * safe way, that reuses most of the code written for the HSS
     * matvec, is to merge y1 and y2 and perform the multiplication
     * with the processes mapped on i. However it means we are
     * performing operations with 0 columns and we do more
     * flops than with the serial code (in the serial code,
     * we do B12*y2 and B21*y1 where y1 and y2 have only
     * nonzero columns. Here [y1;y2] is actually [ext(y1);ext(y2)],
     * i.e., padding with zeros.
     */
    if(myid>=child1->procs[0] && myid<child1->procs[0]+child1->nprow*child1->npcol) {
      /* Mapped on child 1; create tmp = y[ch1] with padded zeros */
      locr=numroc_(&rows_y1,&mb,&nodes[ch1]->myrow,&IZERO,&nodes[ch1]->nprow);
      locc=numroc_(&columns[i],&nb,&nodes[ch1]->mycol,&IZERO,&nodes[ch1]->npcol);
      tmp=new T[locr*locc]();
      idummy=std::max(1,locr);
      descinit_(descPad,&rows_y1,&columns[i],&mb,&nb,&IZERO,&IZERO,&nodes[ch1]->context,&idummy,&ierr);
      idummy=std::max(1,numroc_(&rows_y1,&mb,&child1->myrow,&IZERO,&child1->nprow));
      if(columns[ch1])
        descinit_(descY,&rows_y1,&columns[ch1],&mb,&nb,&IZERO,&IZERO,&child1->context,&idummy,&ierr);
      columns[i]=0;
      columns[ch1]=0;
      for(k=0;k<nJ;k++)
        if(nodes[i]->tree_ib<=J[k] && J[k]<=nodes[i]->tree_ie) {
          columns[i]++;
          if(child1->tree_ib<=J[k] && J[k]<=child1->tree_ie) {
            /* Copy columns corresponding to J[k]: columns[ch1] for child1,
             * columns[i] for tmp (same structure as parent i).
             */
            columns[ch1]++;
            pgeadd('N',rows_y1,IONE,ONE,y[ch1],IONE,columns[ch1],descY,ZERO,tmp,IONE,columns[i],descPad);
           }
        }
    } else if(myid>=child2->procs[0] && myid<child2->procs[0]+child2->nprow*child2->npcol) {
      /* Mapped on child 2; create tmp = y[ch2] with padded zeros */
      locr=numroc_(&rows_y2,&mb,&nodes[ch2]->myrow,&IZERO,&nodes[ch2]->nprow);
      locc=numroc_(&columns[i],&nb,&nodes[ch2]->mycol,&IZERO,&nodes[ch2]->npcol);
      tmp=new T[locr*locc]();
      idummy=std::max(1,locr);
      descinit_(descPad,&rows_y2,&columns[i],&mb,&nb,&IZERO,&IZERO,&nodes[ch2]->context,&idummy,&ierr);
      idummy=std::max(1,numroc_(&rows_y2,&mb,&child2->myrow,&IZERO,&child2->nprow));
      if(columns[ch2])
        descinit_(descY,&rows_y2,&columns[ch2],&mb,&nb,&IZERO,&IZERO,&child2->context,&idummy,&ierr);
      columns[i]=0;
      columns[ch2]=0;
      for(k=0;k<nJ;k++)
        if(nodes[i]->tree_ib<=J[k] && J[k]<=nodes[i]->tree_ie) {
          columns[i]++;
          if(child2->tree_ib<=J[k] && J[k]<=child2->tree_ie) {
            /* Copy columns corresponding to J[k]: columns[ch2] for child2,
             * columns[i] for tmp (same structure as parent i).
             */
            columns[ch2]++;
            pgeadd('N',rows_y2,IONE,ONE,y[ch2],IONE,columns[ch2],descY,ZERO,tmp,IONE,columns[i],descPad);
           }
        }
    } else
      /* Out of both grids */
      tmp=NULL;

    mergeVector(tmp,child1->procs[0],child1->procs[0]+child1->nprow*child1->npcol-1,rows_y1,IONE,child2->procs[0],child2->procs[0]+child2->nprow*child2->npcol-1,rows_y2,IONE,proc_b,proc_g,columns[i],descPad,&tmp2,&desc,nodes[i]->allctxt,&tmpinfo,comm);
    delete[] tmp;

    descY=desc;

    /* Exclude processes that are not in the grid of i */
    if(myid>=proc_b && myid<=proc_g) {

      locr=numroc_(&br,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
      locc=numroc_(&columns[i],&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
      tmp3=new T[locr*locc]();
      idummy=std::max(1,locr);
      descinit_(descPad,&br,&columns[i],&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);

      /* Child 1: z{child1}=B12{i}*y{child2}. y{child2} is in rows rows_y1+1...rows_y1+rows_y2
       * of tmp2, and its columns[ch2] columns are scattered in the columns[i] columns of
       * tmp2. We compute B12 * all the columns and store in tmp3 then we scatter
       * them into the nJ columns of tmp2.
       *  tmp:  [z{child1};z{child2}],               nJ columns.         Descriptor descZ.
       *  tmp2: [y{child1};y{child2}],               columns[i] columns. Descriptor desc.
       *  tmp3: [B12{i}*y{child2},B12{i}*y{child1}], columns[i] columns. Descroptor descPad.
      */
      if(active[ch1]) {
        rows_B=child1->u_rank;
        cols_B=child2->v_rank;
        idummy=std::max(1,numroc_(&rows_B,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow));
        descinit_(descGen,&rows_B,&cols_B,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
        pgemm('N','N',rows_z1,columns[i],rows_y2,ONE,nodes[i]->B12,IONE,IONE,descGen,tmp2,rows_y1+1,IONE,desc,ZERO,tmp3,IONE,IONE,descPad);
      }

      /* Child 2: z{child2}=B21{i}*y{child1}; ditto */
      if(active[ch2]) {
        rows_B=child2->u_rank;
        cols_B=child1->v_rank;
        idummy=std::max(1,numroc_(&rows_B,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow));
        descinit_(descGen,&rows_B,&cols_B,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
        pgemm('N','N',rows_z2,columns[i],rows_y1,ONE,nodes[i]->B21,IONE,IONE,descGen,tmp2,IONE,IONE,desc,ZERO,tmp3,rows_z1+1,IONE,descPad);
      }

      delete[] tmp2;

      /* Scatter the columns[i] columns of tmp3 into the nJ columns of tmp */
      locr=numroc_(&br,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
      locc=numroc_(&nJ,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
      tmp=new T[locr*locc]();
      idummy=std::max(1,locr);
      descZ=new int[BLACSCTXTSIZE];
      descinit_(descZ,&br,&nJ,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);

      columns[i]=0;
      for(k=0;k<nJ;k++)
        if(nodes[i]->tree_ib<=J[k] && J[k]<=nodes[i]->tree_ie) {
          columns[i]++;
          pgeadd('N',br,IONE,ONE,tmp3,IONE,columns[i],descPad,ZERO,tmp,IONE,k+1,descZ);
        }
      delete[] tmp3;

      /* Child 1 and 2: if !root, [z{child1};z{child2}]+=U{i}*z{i} */
      if(i!=dim_tree-1) {
        desc=new int[BLACSCTXTSIZE];
        rows_z=nodes[i]->u_rank;
        idummy=std::max(1,numroc_(&rows_z,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow));
        descinit_(desc,&rows_z,&nJ,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
        nodes[i]->par_applyGenerator('N','U',nJ,ONE,z[i],desc,ONE,tmp,descZ,&tmpinfo);
        delete[] desc;
      }
    } else {
      tmp=NULL;
      descZ=new int[BLACSCTXTSIZE];
      descset_(descZ,&br,&nJ,&mb,&nb,&IZERO,&IZERO,&INONE,&IONE);
    }

    /* Redistribute */
    tmpinfo1.ctxt=child1->context;
    tmpinfo1.nprow=child1->nprow;
    tmpinfo1.npcol=child1->npcol;
    tmpinfo1.myrow=child1->myrow;
    tmpinfo1.mycol=child1->mycol;
    tmpinfo1.mb=ginfo->mb;
    tmpinfo1.nb=ginfo->nb;
    tmpinfo2.ctxt=child2->context;
    tmpinfo2.nprow=child2->nprow;
    tmpinfo2.npcol=child2->npcol;
    tmpinfo2.myrow=child2->myrow;
    tmpinfo2.mycol=child2->mycol;
    tmpinfo2.mb=ginfo->mb;
    tmpinfo2.nb=ginfo->nb;

    splitVector(tmp,child1->procs[0],child1->procs[0]+child1->nprow*child1->npcol-1,rows_z1,child2->procs[0],child2->procs[0]+child2->nprow*child2->npcol-1,rows_z2,nJ,descZ,&tmp2,&desc,nodes[i]->allctxt,&tmpinfo1,&tmpinfo2,comm);
    delete[] tmp;

    if(myid<child1->procs[0]+child1->nprow*child1->npcol) {
      z[ch1]=tmp2;
    } else if(myid>=child2->procs[0] && myid<child2->procs[0]+child2->nprow*child2->npcol) {
      z[ch2]=tmp2;
    }
    delete[] descZ;
    descZ=desc;
  }

  /* Serial subtree */
  if(subtree[2*myid]>=0) {
    istart=subtree[2*myid];
    iend=subtree[2*myid+1];

    for(i=iend;i>=istart;i--) {
      if(!active[i])
        continue; /* Node is not active */
      rows_z=nodes[i]->u_rank;
      if(nodes[i]->nr_children==0) {
        /* Leaf: b(:,:)=U{i}*z{i}+D(:,J) */
        b=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
        tmp=new T[b*nJ]();
        for(j=0;j<nJ;j++)
          if(nodes[i]->tree_ib<=J[j] && J[j]<=nodes[i]->tree_ie)
            lacpy('N',b,IONE,&nodes[i]->D[(J[j]-nodes[i]->tree_ib)*b],b,&tmp[b*j],b);

        if(i!=dim_tree-1)
          /* Avoid weird case leaf=root (1 node in the tree) */
          nodes[i]->seq_applyGenerator('N','U',nJ,ONE,z[i],rows_z,ONE,tmp,b);

        /* The information corresponding to the serial subtree
         * goes into Bl.
         */
        for(j=0;j<nI;j++)
          if(nodes[i]->tree_ib<=I[j] && I[j]<=nodes[i]->tree_ie)
            lacpy('N',IONE,nJ,&tmp[I[j]-nodes[i]->tree_ib],b,&Bl[I[j]-nodes[iend]->tree_ib],ldl);
        delete[] tmp;
      } else {
        /* Non-leaf */
        ch1=children1[i];
        ch2=children2[i];
        child1=nodes[i]->children[0];
        child2=nodes[i]->children[1];
        rows_z1=child1->u_rank;
        rows_z2=child2->u_rank;
        rows_y1=child1->v_rank;
        rows_y2=child2->v_rank;
        /* Child 1: z{child1}=B12{i}*y{child2}+U{i}(1:r1)*z{i}; first part.
         * We first compute B12*y{child2} and scatter the columns[ch2] columns
         * into nJ columns.
         */
        if(active[ch1]) {
          z[ch1]=new T[rows_z1*nJ]();
          if(columns[ch2]) {
            tmp=new T[rows_z1*columns[ch2]]();
            if(child1->u_rank && rows_y2 && rows_z1)
              gemm('N','N',rows_z1,columns[ch2],rows_y2,ONE,nodes[i]->B12,child1->u_rank,y[ch2],rows_y2,ZERO,tmp,rows_z1);
            columns[ch2]=0;
            for(j=0;j<nJ;j++)
              if(child2->tree_ib<=J[j] && J[j]<=child2->tree_ie) {
                lacpy('N',rows_z1,IONE,&tmp[rows_z1*columns[ch2]],rows_z1,&z[ch1][rows_z1*j],rows_z1);
                columns[ch2]++;
              }
            delete[] tmp;
          }
        }

        /* Child 2: z{child2}=B21{i}*y{child1}+U{i}(r1+1:r1+r2)*z{i}; first part
         * We first compute B21*y{child1} and scatter the columns[ch1] columns
         * into nJ columns.
         */
        if(active[ch2]) {
          z[ch2]=new T[rows_z2*nJ]();
          if(columns[ch1]) {
            tmp=new T[rows_z2*columns[ch1]]();
            if(child2->u_rank && rows_y1 && rows_z2)
              gemm('N','N',rows_z2,columns[ch1],rows_y1,ONE,nodes[i]->B21,child2->u_rank,y[ch1],rows_y1,ZERO,tmp,rows_z2);
            columns[ch1]=0;
            for(j=0;j<nJ;j++)
              if(child1->tree_ib<=J[j] && J[j]<=child1->tree_ie) {
                lacpy('N',rows_z2,IONE,&tmp[rows_z2*columns[ch1]],rows_z2,&z[ch2][rows_z2*j],rows_z2);
                columns[ch1]++;
              }
            delete[] tmp;
          }
        }

        /* Child 1 + Child 2; second part */
        if(i!=dim_tree-1) {
          tmp=new T[(rows_z1+rows_z2)*nJ]();
          nodes[i]->seq_applyGenerator('N','U',nJ,ONE,z[i],rows_z,ZERO,tmp,rows_z1+rows_z2);
          for(j=0;j<nJ;j++) {
            if(active[ch1])
              for(k=0;k<rows_z1;k++)
                z[ch1][j*rows_z1+k]+=tmp[j*(rows_z1+rows_z2)+k];
            if(active[ch2])
              for(k=0;k<rows_z2;k++)
                z[ch2][j*rows_z2+k]+=tmp[j*(rows_z1+rows_z2)+k+rows_z1];
          }
          delete[] tmp;
        }
      }
    }
  }

  /* Leaves */
  for(i=0;i<dim_tree;i++) {
    if(   nodes[i]->procs[1]>nodes[i]->procs[0]
       && nodes[i]->nr_children==0
       && nodes[i]->procs[0]+nodes[i]->nprow*nodes[i]->npcol-1>=myid
       && nodes[i]->procs[0]<=myid
       && active[i]
      ) {
      tmpinfo.ctxt=nodes[i]->context;
      tmpinfo.nprow=nodes[i]->nprow;
      tmpinfo.npcol=nodes[i]->npcol;
      tmpinfo.myrow=nodes[i]->myrow;
      tmpinfo.mycol=nodes[i]->mycol;
      tmpinfo.mb=ginfo->mb;
      tmpinfo.nb=ginfo->nb;
      rows_z=nodes[i]->u_rank;
      /* Leaf: b(:,:)=U{i}*z{i}+D(:,J) */
      b=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
      idummy=std::max(1,numroc_(&b,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow));
      descinit_(descGen,&b,&b,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
      for(j=0;j<nJ;j++)
        if(nodes[i]->tree_ib<=J[j] && J[j]<=nodes[i]->tree_ie)
          pgeadd('N',b,IONE,ONE,nodes[i]->D,IONE,J[j]-nodes[i]->tree_ib+1,descGen,ZERO,Bl,IONE,j+1,descBl);
      if(i!=dim_tree-1) {
        idummy=std::max(1,numroc_(&rows_z,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow));
        descZ=new int[BLACSCTXTSIZE];
        descinit_(descZ,&rows_z,&nJ,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
        nodes[i]->par_applyGenerator('N','U',nJ,ONE,z[i],descZ,ONE,Bl,descBl,&tmpinfo);
        delete[] descZ;
      }
    }
  }

  /* Combine rows (Bl) into the ouput B */
  extractRowVectorTree(Bl,descBl,I,nI,nJ,B,descB,ginfo,nodes[dim_tree-1]->allctxt);
  delete[] Bl;

  for(i=0;i<dim_tree-1;i++) {
    delete[] y[i];
    delete[] z[i];
  }
  delete[] y;
  delete[] z;

}

#endif // HSS_PAR_EXTRACTION_HPP
