/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

#ifndef SEQ_HQRINTERPOL_HPP
#define SEQ_HQRINTERPOL_HPP

#include "../TaskTimer.hpp"

template<typename T, typename S> void seq_HQRInterpol(int m, int n, T *A, S tol, T **X, int *r, int **J, int **P) {
  /* Serial Interpolative Decomposition of A.
   * Decompose the m x n matrix A into A=A(:,J) [I X]*P^T
   * with J a set of columns of size r, and X rk x (n-rk).
   * Step 1: RRQR(A) => A P = Q R = Q [R1 R2]
   * Step 2: TRSM: A P = (Q R1) (R1^-1 R) => A = A(:,J)*[I X]*P^T
   *
   * Matrices are assumed to be stored in column-major format.
   * Matrix A is NOT modified.
   *
   * Matrix Q is initially a copy of A, then it contains the Q factor and is eventually discarded.
   * Matrix X has initial size n x n but is eventually reshaped into r x (n-r) (0-rows and identity columns are discarded).
   *
   * This is a block-Householder algorithm. It is a essentially a modified version of LAPACK's
   * xgeqp3 with a threshold.
   */

  TIMER_TIME(HSS_SEQHQRINTERPOL, 1, t_hss_seq_hqr);

  int i, j;
  S *norms;
  T *Q, *R;

  T ZERO=0;
  T  ONE=1;

#if defined(checkID)
  T *B, *C;
  S *work;
#endif

  /* Q: m x n, copy of A */
  Q=new T[m*n];
  for(i=0;i<m*n;i++)
    Q[i]=A[i];

  /* R: n x n */
  R=new T[n*n];
  for(i=0;i<n*n;i++)
    R[i]=ZERO;

  /* J: indices of permuted colums */
  *J=new int[n];
  for(i=0;i<n;i++)
      (*J)[i]=0; /* 0: free columns for geqp3 */

  /* P: permutation of columns */
  *P=new int[n];
  for(i=0;i<n;i++)
      (*P)[i]=i+1;

  /* norms of columns */
  norms=new S[n];

  /* Step 1: RRQR */
  *r=0;
#if defined (DGEQP4)
  geqp4mod(m,n,Q,*J,r,tol);
#else
  geqp3mod(m,n,Q,*J,r,tol);
#endif

  /* Copy the R factor into R */
  lacpy('U',*r,n,Q,m,R,n);

  /* Generate P from J */
  if(*r)
    for(i=1;i<=n;i++) {
      j=(*J)[i-1];
      while(j<i)
        j=(*J)[j-1];
      (*P)[i-1]=j;
    }

#if defined(debug)
  std::cout << "Rank = " << *r << std::endl << std::endl;
  printMatrix("Q",Q,m,n,m);
  printMatrix("R",R,n,n,n);
#endif

  delete[] Q;
  delete[] norms;

  if(*r==0) {
    delete[] R;
    *X=NULL;
    delete[] *J;
    *J=NULL;
    return;
  }

  /* Step 2: TRSM and permutation: R1^-1 R = [I R1^-1 R2] = [I X] with R = [R1 R2], R1 r x r */
  trsm('L','U','N','N',*r,n-*r,ONE,R,n,&R[*r*n],n);

  *X=new T[*r*(n-*r)];
  for(j=0;j<n-*r;j++)
    for(i=0;i<*r;i++)
      (*X)[j*(*r)+i]=R[(*r+j)*n+i];
  delete[] R;

#if defined(checkID)
  /* Copy A into B. Compute C=A(:,J)*[I X] by copying A(:,J) into C(:,1:r) then
   * computing C(:,r+1:n)=C(:,1:r)*X. Then permute into C=C P^T.
   * Finally compute B=B-C=A-A(:,J)*[I X]*P^T and norm(B)/norm(A). */
  B=new T[m*n];
  for(i=0;i<m*n;i++)
    B[i]=A[i];

  C=new T[m*n];
  for(i=0;i<*r;i++)
    lacpy('N',m,1,&A[((*J)[i]-1)*m],m,&C[i*m],m);
  gemm('N','N',m,n-*r,*r,ONE,C,m,*X,*r,ZERO,&C[*r*m],m);
  lapmt(IZERO,m,n,C,m,*J);

  for(i=0;i<m*n;i++)
    B[i]=B[i]-C[i];

  work=new S[m];
  std::cout << "Serial ID: ||A-A(:,J)X||_inf/||A||_inf=    " << lange('I',m,n,B,m,work)/lange('I',m,n,A,m,work) << std::endl << std::endl;
  delete[] B;
  delete[] C;
  delete[] work;
#endif

  int *Jtmp=new int[*r];
  for(int i=0;i<*r;i++)
    Jtmp[i]=(*J)[i];
  delete[] *J;
  *J=Jtmp;

#if defined(debug)
  printMatrix("X",*X,*r,n-*r,*r);
  printMatrix("J",*J,1,*r,1);
  printMatrix("P",*J,1,*r,1);
#endif

}

#endif // SEQ_HQRINTERPOL_HPP
