/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Pieter Ghysels, Francois-Henry Rouet, Xiaoye S. Li.
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */
#ifndef FRONTAL_MATRIX_HSS_HPP
#define FRONTAL_MATRIX_HSS_HPP

#include <iostream>
#include <algorithm>
#include <memory>
#include <complex>
#include <algorithm>
#include "FrontalMatrix.hpp"
#include "blas_lapack_wrapper.hpp"
#include "CompressedSparseMatrix.hpp"
#include "MatrixReordering.hpp"
#include "HSSMatrix.hpp"
#include "TaskTimer.hpp"
#include "Random_wrapper.hpp"

namespace strumpack {

  // forward declaration
  template<typename scalar_t,typename real_t,typename integer_t> class HSSMatrix;
  template<typename scalar_t,typename real_t,typename integer_t> class FrontalMatrixDense;

  template<typename scalar_t,typename real_t,typename integer_t> class FrontalMatrixHSS
    : public FrontalMatrix<scalar_t,real_t,integer_t> {
  public:
    /** TODO why do we need this?? why is it passed to HSS compression routine?? */
    integer_t* row_ind;
    integer_t* col_ind;
    HSSMatrix<scalar_t,real_t,integer_t>* hssmat;

    /** random vectors and samples */
    scalar_t* R;
    scalar_t* Srow;
    scalar_t* Scol;
    scalar_t* Rrow;
    scalar_t* Rcol;
    /** number of random vectors */
    integer_t nr_random_vectors;

    /**
     * this is the low rank Schur complement update
     *   S = F22 - theta * Vhat^* * phi
     * where
     * dim_reduced_2 = hssmat->child[0]->u_rows
     * dim_reduced_1 = hssmat->child[0]->v_rank
     *  - F22    is dim_upd x dim_upd
     *  - theta  is dim_upd x dim_reduced_1
     *  - Vhat^* is dim_reduced_1 x dim_reduced_2
     *  - phi is dim_reduced_2 x dim_upd (phi is stored transposed!!)
     * theta and phi are allocated here, then computed using the hssmat.
     * Vhat is stored in the hssmat with ld dim_reduced_2 = hssmat->child[0]->u_rows
     */
    integer_t dim_reduced_1;
    integer_t dim_reduced_2;
    scalar_t* theta;
    scalar_t* phi;
    scalar_t* Vhat;

    FrontalMatrixHSS(CompressedSparseMatrix<scalar_t,real_t,integer_t>* _A,
		     integer_t _sep, integer_t _sep_begin, integer_t _sep_end,
		     integer_t _dim_upd, integer_t* _upd);
    ~FrontalMatrixHSS();
    void merge_random_vectors_from_children(int task_depth);
    void extend_add(int task_depth) {}
    void random_sampling(std::unique_ptr<RandomGeneratorBase<real_t>>& rgen, HSSStrategy<real_t>& strat, int task_depth);
    void add_more_random_vectors(std::unique_ptr<RandomGeneratorBase<real_t>>& rgen, int task_depth);
    void fill_random(std::unique_ptr<RandomGeneratorBase<real_t>>& rgen, integer_t cols, scalar_t* Z, int task_depth);
    void multiply_random(integer_t cols, scalar_t* _R, scalar_t* _Srow, scalar_t* _Scol, int task_depth);
    void UUtxR(integer_t cols, scalar_t* newR, scalar_t* newSrow, scalar_t* newScol,
	       FrontalMatrix<scalar_t,real_t,integer_t>* parent, int task_depth);
    void release_work_memory();
    void multifrontal_factorization(HSSStrategy<real_t>& strat, int etree_level=0, int task_depth=0);
    void forward_multifrontal_solve(scalar_t* b, scalar_t* wmem, int etree_level=0, int task_depth=0);
    void backward_multifrontal_solve(scalar_t* y, scalar_t* wmem, int etree_level=0, int task_depth=0);
    void fast_Schur_update(int task_depth);
    void get_submatrix(integer_t* rows, integer_t nr_rows, integer_t* cols, integer_t nr_cols,
		       scalar_t* values, integer_t ld, int task_depth);
    integer_t get_maximum_rank(int task_depth=0) const;
    void print_rank_statistics(std::ostream &out);
    bool isHSS() { return true; };
    std::string type() { return "FrontalMatrixHSS"; }
    void set_HSS_partitioning(HSSPartitionTree& tree, bool is_root);
#if defined(dbgHSS)
    FrontalMatrixDense<scalar_t,real_t,integer_t>* df;
    FrontalMatrix<scalar_t,real_t,integer_t>* get_df() { return df; }
#endif

  private:
    FrontalMatrixHSS(const FrontalMatrixHSS&) = delete;
    FrontalMatrixHSS& operator=(FrontalMatrixHSS const&) = delete;
    void multifrontal_factorization_node(HSSStrategy<real_t>& strat, int etree_level, int task_depth);
    void forward_multifrontal_solve_node(scalar_t* b, scalar_t* wmem, int etree_level, int task_depth);
    void backward_multifrontal_solve_node(scalar_t* y, scalar_t* wmem, int etree_level, int task_depth);
    long long get_node_factor_nonzeros();
    long long get_dense_node_factor_nonzeros();
  };

  template<typename scalar_t,typename real_t,typename integer_t>
  FrontalMatrixHSS<scalar_t,real_t,integer_t>::FrontalMatrixHSS
  (CompressedSparseMatrix<scalar_t,real_t,integer_t>* _A, integer_t _sep,
   integer_t _sep_begin, integer_t _sep_end, integer_t _dim_upd, integer_t* _upd)
    : FrontalMatrix<scalar_t,real_t,integer_t>
    (_A, NULL, NULL, _sep, _sep_begin, _sep_end, _dim_upd, _upd),
    hssmat(NULL), R(NULL), Srow(NULL), Scol(NULL), Rrow(NULL), Rcol(NULL),
    dim_reduced_1(0), dim_reduced_2(0), theta(NULL), phi(NULL), Vhat(NULL) {
    row_ind = new integer_t[2*this->dim_blk];
    col_ind = &row_ind[this->dim_blk];
    for (integer_t i=0; i<this->dim_sep; i++) row_ind[i] = col_ind[i] = this->sep_begin + i;
    for (integer_t i=0; i<this->dim_upd; i++) row_ind[this->dim_sep + i] = col_ind[this->dim_sep + i] = this->upd[i];
#if defined(dbgHSS)
    df = new FrontalMatrixDense<scalar_t,real_t,integer_t>(_A, _sep, _sep_begin, _sep_end, _dim_upd, _upd);
#endif
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixHSS<scalar_t,real_t,integer_t>::set_HSS_partitioning(HSSPartitionTree& tree, bool is_root) {
    assert(tree.size == this->dim_sep);
    if (is_root) hssmat = new HSSMatrix<scalar_t,real_t,integer_t>(tree, 0);
    else hssmat = new HSSMatrix<scalar_t,real_t,integer_t>(tree, this->dim_upd, 0);
  }

  template<typename scalar_t,typename real_t,typename integer_t>
  FrontalMatrixHSS<scalar_t,real_t,integer_t>::~FrontalMatrixHSS() {
    delete[] R;
    delete[] theta;
    delete hssmat;
    delete[] row_ind;
#if defined(dbgHSS)
    if (df) {
      df->lchild = NULL;
      df->rchild = NULL;
      delete[] df->F22; df->F22 = NULL;
      delete df;
      df = NULL;
    }
#endif
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixHSS<scalar_t,real_t,integer_t>::get_submatrix
  (integer_t* rows, integer_t nr_rows, integer_t* cols, integer_t nr_cols,
   scalar_t* values, integer_t ld, int task_depth) {
    std::vector<integer_t> my_cols;    my_cols.reserve(nr_cols);
    std::vector<integer_t> orig_cols;  orig_cols.reserve(nr_cols);
    for (integer_t i=0; i<nr_cols; i++) {
      integer_t* l = std::lower_bound(this->upd, this->upd+this->dim_upd, cols[i]);
      if (l != this->upd+this->dim_upd && *l == cols[i]) {
	my_cols.push_back((l-this->upd)+this->dim_sep);
	orig_cols.push_back(i);
      }
    }
    if (my_cols.empty()) return;
    std::vector<integer_t> my_rows;    my_rows.reserve(nr_rows);
    std::vector<integer_t> orig_rows;  orig_rows.reserve(nr_rows);
    for (integer_t i=0; i<nr_rows; i++) {
      integer_t* l = std::lower_bound(this->upd, this->upd+this->dim_upd, rows[i]);
      if (l != this->upd+this->dim_upd && *l == rows[i]) {
	my_rows.push_back((l-this->upd)+this->dim_sep);
	orig_rows.push_back(i);
      }
    }
    if (my_rows.empty()) return;
    // TODO can extract from just the 22 block, ie child[1], if not
    // the root (cannot be root, no extraction from root)
    hssmat->extract_submatrix(my_rows, my_cols, orig_rows, orig_cols, values, ld,
			      this->dim_upd, this->dim_sep, task_depth);
    if (dim_reduced_1 && dim_reduced_2) {
      /**
       * extract rows of theta and put them as columns in work1 (transpose later)
       * extract rows of phi and put them in work2
       *  (note that phi is stored transposed, so extracting rows iso columns)
       * compute Vhat^* x work2, put result in work3
       * compute work1 x work3, put result in values
       */
      auto work1 = new scalar_t[dim_reduced_1 * nr_rows + dim_reduced_2 * nr_cols + dim_reduced_1 * nr_cols];
      auto work2 = work1 + dim_reduced_1 * nr_rows;
      auto work3 = work2 + dim_reduced_2 * nr_cols;
      std::fill(work1, work1 + dim_reduced_1 * nr_rows, scalar_t(0.));
      std::fill(work2, work2 + dim_reduced_2 * nr_cols, scalar_t(0.));
      for (size_t r=0; r<my_rows.size(); r++)
	blas::copy(dim_reduced_1, theta+my_rows[r]-this->dim_sep, this->dim_upd, work1+orig_rows[r]*dim_reduced_1, 1);
      for (size_t c=0; c<my_cols.size(); c++) {
	// this triggers an invalid read in valgrind ?
	//blas::copy(dim_reduced_2, phi+my_cols[c]-this->dim_sep, this->dim_upd, work2+orig_cols[c]*dim_reduced_2, 1);
	for (int i=0; i<dim_reduced_2; i++)
	  work2[orig_cols[c]*dim_reduced_2 + i] = phi[my_cols[c]-this->dim_sep +i*this->dim_upd];
	blas::lacgv(dim_reduced_2, work2+orig_cols[c]*dim_reduced_2, 1);
      }
      gemm_omp_task('C', 'N', dim_reduced_1, nr_cols, dim_reduced_2, scalar_t(1.),
		    Vhat, dim_reduced_2, work2, dim_reduced_2, scalar_t(0.), work3, dim_reduced_1, task_depth);
      gemm_omp_task('T', 'N', nr_rows, nr_cols, dim_reduced_1, scalar_t(-1.),
		    work1, dim_reduced_1, work3, dim_reduced_1, scalar_t(1.), values, ld, task_depth);
      delete[] work1;
    }
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixHSS<scalar_t,real_t,integer_t>::release_work_memory() {
    delete[] R; R=NULL; Srow=NULL; Scol=NULL; Rrow=NULL; Rcol=NULL;
    delete[] row_ind; row_ind=col_ind=NULL;
    hssmat->destroy_update_block();
#if defined(dbgHSS)
    if (df) {
      df->lchild = NULL;
      df->rchild = NULL;
      delete[] df->F22; df->F22 = NULL;
      delete df;
      df = NULL;
    }
#endif
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixHSS<scalar_t,real_t,integer_t>::fast_Schur_update(int task_depth) {
    dim_reduced_1 = hssmat->child[0]->v_rank;
    dim_reduced_2 = hssmat->child[0]->u_rows;
    if (dim_reduced_1 && dim_reduced_2) {   // U_i = F_22 - theta phi^T
      theta = new scalar_t[(dim_reduced_1+dim_reduced_2)*this->dim_upd];
      std::fill(theta, theta+(dim_reduced_1+dim_reduced_2)*this->dim_upd, scalar_t(0.));
      phi = &theta[this->dim_upd*dim_reduced_1];
      hssmat->fast_Schur_update(theta-this->dim_sep, this->dim_upd, phi-this->dim_sep, this->dim_upd, task_depth);
      Vhat = hssmat->child[0]->Vt;
    }
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixHSS<scalar_t,real_t,integer_t>::UUtxR
  (integer_t cols, scalar_t* newR, scalar_t* newSrow, scalar_t* newScol,
   FrontalMatrix<scalar_t,real_t,integer_t>* parent, int task_depth) {
    if (!this->dim_upd) return;
    auto ldRtmp = this->dim_upd;
    auto ldStmp = this->dim_upd;
    auto Rtmp = new scalar_t[2*cols*this->dim_upd + dim_reduced_2*cols+dim_reduced_1*cols]();
    auto Stmp = Rtmp + cols*this->dim_upd;
    auto work1 = Stmp + cols*this->dim_upd;
    auto work2 = work1 + dim_reduced_2*cols;
    this->extract_child_rows(cols, newR, Rtmp, parent, task_depth);
    if (dim_reduced_1 && dim_reduced_2) {   // Srow = CB*R
      gemm_omp_task('C', 'N', dim_reduced_2, cols, this->dim_upd, scalar_t(1.), phi,
		    this->dim_upd, Rtmp, ldRtmp, scalar_t(0.), work1, dim_reduced_2, task_depth);
      gemm_omp_task('C', 'N', dim_reduced_1, cols, dim_reduced_2, scalar_t(1.), Vhat,
		    dim_reduced_2, work1, dim_reduced_2, scalar_t(0.), work2, dim_reduced_1, task_depth);
      gemm_omp_task('N', 'N', this->dim_upd, cols, dim_reduced_1, scalar_t(-1.), theta,
		    this->dim_upd, work2, dim_reduced_1, scalar_t(0.), Stmp, ldStmp, task_depth);
    }
    hssmat->child[1]->matvec(Rtmp-this->dim_sep, ldRtmp, scalar_t(1.),
			     Stmp-this->dim_sep, ldStmp, cols, task_depth);
    this->skinny_extend_add(cols, Stmp, parent, newSrow, task_depth);
    if (dim_reduced_1 && dim_reduced_2) {   // Scol = CB^T*R
      gemm_omp_task('C', 'N', dim_reduced_1, cols, this->dim_upd, scalar_t(1.), theta,
		    this->dim_upd, Rtmp, ldRtmp, scalar_t(0.), work2, dim_reduced_1, task_depth);
      gemm_omp_task('N', 'N', dim_reduced_2, cols, dim_reduced_1, scalar_t(1.), Vhat,
		    dim_reduced_2, work2, dim_reduced_1, scalar_t(0.), work1, dim_reduced_2, task_depth);
      gemm_omp_task('N', 'N', this->dim_upd, cols, dim_reduced_2, scalar_t(-1.), phi,
		    this->dim_upd, work1, dim_reduced_2, scalar_t(0.), Stmp, ldStmp, task_depth);
    }
    hssmat->child[1]->matvec_transpose(Rtmp-this->dim_sep, ldRtmp, scalar_t(1.),
				       Stmp-this->dim_sep, ldStmp, cols, task_depth);
    this->skinny_extend_add(cols, Stmp, parent, newScol, task_depth);
    delete[] Rtmp;
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixHSS<scalar_t,real_t,integer_t>::multiply_random
  (integer_t cols, scalar_t* _R, scalar_t* _Srow, scalar_t* _Scol, int task_depth) {
    const size_t N = cols * this->dim_blk;
    const integer_t B = params::random_vector_panel;
    std::fill(_Srow, _Srow+N, scalar_t(0.));
    std::fill(_Scol, _Scol+N, scalar_t(0.));
    for (integer_t c=0; c<cols; c+=B)
#pragma omp task default(shared) firstprivate(c) if(task_depth < params::task_recursion_cutoff_level)
      {
	integer_t o = c*this->dim_blk;
	this->A->front_multiply(this->sep_begin, this->sep_end, this->upd, this->dim_sep, this->dim_upd,
				std::min(B,cols-c), _R+o, _Srow+o, _Scol+o, this->dim_blk);
      }
#pragma omp taskwait
    if (this->lchild) this->lchild->UUtxR(cols, _R, _Srow, _Scol, this, task_depth);
    if (this->rchild) this->rchild->UUtxR(cols, _R, _Srow, _Scol, this, task_depth);
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixHSS<scalar_t,real_t,integer_t>::random_sampling
  (std::unique_ptr<RandomGeneratorBase<real_t>>& rgen, HSSStrategy<real_t>& strat, int task_depth) {
    nr_random_vectors = strat.get_nr_random_vectors(this->dim_sep);
    auto N = nr_random_vectors * this->dim_blk;
    R = new scalar_t[5*N];
    Srow = R + N;
    Scol = R + 2*N;
    Rrow = R + 3*N;
    Rcol = R + 4*N;
    fill_random(rgen, nr_random_vectors, R, task_depth);
    std::copy(R, R+N, Rrow);
    std::copy(R, R+N, Rcol);
    multiply_random(nr_random_vectors, R, Srow, Scol, task_depth);
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixHSS<scalar_t,real_t,integer_t>::add_more_random_vectors
  (std::unique_ptr<RandomGeneratorBase<real_t>>& rgen, int task_depth) {
    auto o = nr_random_vectors * this->dim_blk;
    auto N = 2 * o;
    auto tmp = new scalar_t[5*N];
    std::copy(R, R+o, tmp);
    std::copy(Srow, Srow+o, tmp+N);   Srow = tmp + N;
    std::copy(Scol, Scol+o, tmp+2*N); Scol = tmp + 2*N;
    std::copy(Rrow, Rrow+o, tmp+3*N); Rrow = tmp + 3*N;
    std::copy(Rcol, Rcol+o, tmp+4*N); Rcol = tmp + 4*N;
    delete[] R;
    R = tmp;
    fill_random(rgen, nr_random_vectors, R+o, task_depth);
    std::copy(R+o, R+N, Rrow+o);
    std::copy(R+o, R+N, Rcol+o);
    multiply_random(nr_random_vectors, R+o, Srow+o, Scol+o, task_depth);
    nr_random_vectors *= 2;
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixHSS<scalar_t,real_t,integer_t>::fill_random
  (std::unique_ptr<RandomGeneratorBase<real_t>>& rgen, integer_t cols, scalar_t* Z, int task_depth) {
    integer_t N = cols * this->dim_blk;
    for (integer_t i=0; i<N; i++) Z[i] = rgen->get();
    STRUMPACK_FLOPS(static_cast<long long int>(N) * rgen->flops_per_prng());
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixHSS<scalar_t,real_t,integer_t>::multifrontal_factorization
  (HSSStrategy<real_t>& strat, int etree_level, int task_depth) {
    if (task_depth == 0)
#pragma omp parallel if(!omp_in_parallel())
#pragma omp single
      multifrontal_factorization_node(strat, etree_level, task_depth);
    else multifrontal_factorization_node(strat, etree_level, task_depth);
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixHSS<scalar_t,real_t,integer_t>::multifrontal_factorization_node
  (HSSStrategy<real_t>& strat, int etree_level, int task_depth) {
    bool tasked = task_depth < params::task_recursion_cutoff_level;
    if (tasked) {
      if (this->lchild)
#pragma omp task untied default(shared) final(task_depth >= params::task_recursion_cutoff_level-1) mergeable
	this->lchild->multifrontal_factorization(strat, etree_level+1, task_depth+1);
      if (this->rchild)
#pragma omp task untied default(shared) final(task_depth >= params::task_recursion_cutoff_level-1) mergeable
	this->rchild->multifrontal_factorization(strat, etree_level+1, task_depth+1);
#pragma omp taskwait
    } else {
      if (this->lchild) this->lchild->multifrontal_factorization(strat, etree_level+1, task_depth);
      if (this->rchild) this->rchild->multifrontal_factorization(strat, etree_level+1, task_depth);
    }

#if defined(dbgHSS)
    scalar_t* fe = new scalar_t[df->dim_blk*df->dim_blk];
    df->lchild = (this->lchild) ? this->lchild->get_df() : NULL;
    df->rchild = (this->rchild) ? this->rchild->get_df() : NULL;
    df->F11 = new scalar_t[df->dim_sep*df->dim_sep+2*df->dim_sep*df->dim_upd];
    df->F12 = df->F11+df->dim_sep*df->dim_sep;
    df->F21 = df->F12+df->dim_upd*df->dim_sep;
    df->piv = new int[df->dim_sep];
    std::fill(df->F11, df->F11+df->dim_sep*df->dim_sep+2*df->dim_sep*df->dim_upd, scalar_t(0.));
    std::fill(df->piv, df->piv+df->dim_sep, 0);
    df->A->extract_front(df->F11, df->F12, df->F21, df->dim_sep, df->dim_upd,
			 df->sep_begin, df->sep_end, df->upd, task_depth);
    if (df->dim_upd) {
      df->F22 = new scalar_t[df->dim_upd*df->dim_upd];
      std::fill(df->F22, df->F22+df->dim_upd*df->dim_upd, scalar_t(0.));
    }
    if (df->lchild) df->extend_add(static_cast<FrontalMatrixDense<scalar_t,real_t,integer_t>* >(df->lchild), task_depth);
    if (df->rchild) df->extend_add(static_cast<FrontalMatrixDense<scalar_t,real_t,integer_t>* >(df->rchild), task_depth);
    blas::omatcopy('N', df->dim_sep, df->dim_sep, df->F11, df->dim_sep, fe, df->dim_blk);
    blas::omatcopy('N', df->dim_sep, df->dim_upd, df->F12, df->dim_sep, fe+df->dim_sep*df->dim_blk, df->dim_blk);
    blas::omatcopy('N', df->dim_upd, df->dim_sep, df->F21, df->dim_upd, fe+df->dim_sep, df->dim_blk);
    blas::omatcopy('N', df->dim_upd, df->dim_upd, df->F22, df->dim_upd, fe+df->dim_sep*df->dim_blk+df->dim_sep, df->dim_blk);
    df->multifrontal_factorization_phase2(strat, etree_level, task_depth, tasked);
#endif

    auto rgen = make_random_generator<real_t>();
    random_sampling(rgen, strat, task_depth);
    do {
      hssmat->compress(this, strat, nr_random_vectors, Rrow, Rcol, this->dim_blk, Scol, Srow, this->dim_blk, task_depth, true);
      if (hssmat->state != HSSMatrix<scalar_t,real_t,integer_t>::COMPRESSED) {
	// this is overwritten during compression so needs to be set again
	// maybe there is a cleaner solution than this
	for (integer_t i=0; i<this->dim_sep; i++) row_ind[i] = col_ind[i] = this->sep_begin + i;
	for (integer_t i=0; i<this->dim_upd; i++) row_ind[this->dim_sep + i] = col_ind[this->dim_sep + i] = this->upd[i];
	add_more_random_vectors(rgen, tasked);
      }
    } while (hssmat->state != HSSMatrix<scalar_t,real_t,integer_t>::COMPRESSED);

    if (params::log_ranks) {
      auto N = this->dim_blk*this->dim_blk;
      scalar_t* tmp = new scalar_t[N];
      for (int c=0; c<this->dim_blk; c++)
      	for (int r=0; r<this->dim_blk; r++)
	  tmp[r+c*this->dim_blk] = (r == c) ? scalar_t(1.) : scalar_t(0.);
      scalar_t* front = new scalar_t[N];
      hssmat->matvec(tmp, this->dim_blk, scalar_t(0.), front, this->dim_blk, this->dim_blk, task_depth);
      real_t nrm_front = blas::nrm2<scalar_t,real_t>(N, front, 1);
#if defined(dbgHSS)
      blas::axpy(N, scalar_t(-1.), front, 1, fe, 1);
      real_t acc_total = blas::nrm2<scalar_t,real_t>(N, fe, 1) / nrm_front;
      delete[] fe;
#else
      real_t acc_total = -1.;
#endif
      for (int c=0; c<this->dim_blk; c++)
	for (int r=0; r<this->dim_blk; r++)
	  tmp[r+c*this->dim_blk] = scalar_t(0.);
      auto rows = new integer_t[2*this->dim_blk];
      auto cols = rows + this->dim_blk;
      for (int i=0; i<this->dim_sep; i++) rows[i] = cols[i] = i+this->sep_begin;
      for (int i=0; i<this->dim_upd; i++) rows[i+this->dim_sep] = cols[i+this->dim_sep] = this->upd[i];
      this->extract(rows, this->dim_blk, cols, this->dim_blk, tmp, this->dim_blk, task_depth);
      delete[] rows;
      blas::axpy(N, scalar_t(-1.), tmp, 1, front, 1);
      delete[] tmp;
      real_t acc_front = blas::nrm2<scalar_t,real_t>(N, front, 1) / nrm_front;
      delete[] front;
      std::fstream fs("rank_stats_" + std::to_string(mpi_rank()) + ".log", std::fstream::out | std::fstream::app);
      fs << "SEQ   \t " << this->sep << " \t " << this->dim_sep << " \t " << this->dim_upd
	 << " \t " << nr_random_vectors << " \t " << hssmat->get_maximum_rank()
	 << "\tapp err = " << acc_front << "\tcumm err = " << acc_total
	 << "\t[" << this->sep_begin << "," << this->sep_end << "], ";
      if (this->lchild && this->rchild)
	fs << "lch=[" << this->lchild->sep_begin << "," << this->lchild->sep_end << "]/" << this->lchild->type()
	   << " rch=[" << this->rchild->sep_begin << "," << this->rchild->sep_end << "]/" << this->rchild->type();
      fs << std::endl;
      fs.close();
    }

    if (this->lchild) this->lchild->release_work_memory();
    if (this->rchild) this->rchild->release_work_memory();

    if (this->dim_sep) {
      if (etree_level > 0) {
	hssmat->child[0]->ULV_factorization(task_depth, true);
	fast_Schur_update(task_depth);
      } else hssmat->ULV_factorization(task_depth, true);
    }
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixHSS<scalar_t,real_t,integer_t>::forward_multifrontal_solve
  (scalar_t* b, scalar_t* wmem, int etree_level, int task_depth) {
    if (task_depth == 0)
#pragma omp parallel if(!omp_in_parallel())
#pragma omp single
      forward_multifrontal_solve_node(b, wmem, etree_level, task_depth);
    else forward_multifrontal_solve_node(b, wmem, etree_level, task_depth);
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixHSS<scalar_t,real_t,integer_t>::forward_multifrontal_solve_node
  (scalar_t* b, scalar_t* wmem, int etree_level, int task_depth) {
    bool tasked = task_depth<params::task_recursion_cutoff_level;
    if (tasked) {
      if (this->lchild)
#pragma omp task untied default(shared) final(task_depth >= params::task_recursion_cutoff_level-1) mergeable
	this->lchild->forward_multifrontal_solve(b, wmem, etree_level+1, task_depth+1);
      if (this->rchild)
#pragma omp task untied default(shared) final(task_depth >= params::task_recursion_cutoff_level-1) mergeable
	this->rchild->forward_multifrontal_solve(b, wmem, etree_level+1, task_depth+1);
#pragma omp taskwait
    } else {
      if (this->lchild) this->lchild->forward_multifrontal_solve(b, wmem, etree_level+1, task_depth);
      if (this->rchild) this->rchild->forward_multifrontal_solve(b, wmem, etree_level+1, task_depth);
    }
    this->look_left(b, wmem);
    if (etree_level > 0) {
      if (dim_reduced_1 && dim_reduced_2) {
	auto reduced_rhs = new scalar_t[dim_reduced_1];
	hssmat->child[0]->ULV_forward_solve(b+this->sep_begin, reduced_rhs, task_depth, true);
	// reduced_rhs = \hat{V}^* y_0 + V^* [z_0; z_1]
	if (this->dim_upd)
	  gemv_omp_task('N', this->dim_upd, dim_reduced_1, scalar_t(-1.), theta, this->dim_upd,
			reduced_rhs, 1, scalar_t(1.), wmem + this->p_wmem, 1, task_depth);
	delete[] reduced_rhs;
      }
    } else hssmat->ULV_forward_solve(b+this->sep_begin, NULL, task_depth, true);
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixHSS<scalar_t,real_t,integer_t>::backward_multifrontal_solve
  (scalar_t* b, scalar_t* wmem, int etree_level, int task_depth) {
    if (task_depth == 0)
#pragma omp parallel if(!omp_in_parallel())
#pragma omp single
      backward_multifrontal_solve_node(b, wmem, etree_level, task_depth);
    else backward_multifrontal_solve_node(b, wmem, etree_level, task_depth);
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixHSS<scalar_t,real_t,integer_t>::backward_multifrontal_solve_node
  (scalar_t* y, scalar_t* wmem, int etree_level, int task_depth) {
    bool tasked = task_depth<params::task_recursion_cutoff_level;
    if (etree_level > 0) {
      if (dim_reduced_1 && dim_reduced_2) {
	if (this->dim_upd)
	  gemv_omp_task('C', this->dim_upd, dim_reduced_2, scalar_t(-1.), phi, this->dim_upd,
			wmem+this->p_wmem, 1, scalar_t(1.), hssmat->child[0]->work_u, 1, task_depth);
	hssmat->child[0]->ULV_backward_solve(y+this->sep_begin, task_depth, true);
      }
    } else hssmat->ULV_backward_solve(y+this->sep_begin, task_depth, true);
    this->look_right(y, wmem);
    if (tasked) {
      if (this->lchild)
#pragma omp task untied default(shared) final(task_depth >= params::task_recursion_cutoff_level-1) mergeable
	this->lchild->backward_multifrontal_solve(y, wmem, etree_level+1, task_depth+1);
      if (this->rchild)
#pragma omp task untied default(shared) final(task_depth >= params::task_recursion_cutoff_level-1) mergeable
	this->rchild->backward_multifrontal_solve(y, wmem, etree_level+1, task_depth+1);
#pragma omp taskwait
    } else {
      if (this->lchild) this->lchild->backward_multifrontal_solve(y, wmem, etree_level+1, task_depth);
      if (this->rchild) this->rchild->backward_multifrontal_solve(y, wmem, etree_level+1, task_depth);
    }
  }

  template<typename scalar_t,typename real_t,typename integer_t> integer_t
  FrontalMatrixHSS<scalar_t,real_t,integer_t>::get_maximum_rank(int task_depth) const {
    integer_t max_rank = hssmat->get_maximum_rank();
    integer_t max_rank_l = 0;
    integer_t max_rank_r = 0;
    if (this->lchild)
#pragma omp task untied default(shared) final(task_depth >= params::task_recursion_cutoff_level-1) mergeable
      max_rank_l = this->lchild->get_maximum_rank(task_depth+1);
    if (this->rchild)
#pragma omp task untied default(shared) final(task_depth >= params::task_recursion_cutoff_level-1) mergeable
      max_rank_r = this->rchild->get_maximum_rank(task_depth+1);
#pragma omp taskwait
    return std::max(max_rank, std::max(max_rank_l, max_rank_r));
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixHSS<scalar_t,real_t,integer_t>::print_rank_statistics(std::ostream &out) {
    if (this->lchild) this->lchild->print_rank_statistics(out);
    if (this->rchild) this->rchild->print_rank_statistics(out);
    hssmat->print_ranks(out);
  }

  template<typename scalar_t,typename real_t,typename integer_t> long long
  FrontalMatrixHSS<scalar_t,real_t,integer_t>::get_node_factor_nonzeros() {
    long long nnz = (dim_reduced_1 + dim_reduced_2) * this->dim_upd;  // theta + phi
    return nnz + hssmat->get_factor_nonzeros();
  }

  template<typename scalar_t,typename real_t,typename integer_t> long long
  FrontalMatrixHSS<scalar_t,real_t,integer_t>::get_dense_node_factor_nonzeros() {
    return this->dim_blk*this->dim_blk-this->dim_upd*this->dim_upd;
  }

} // end namespace strumpack

#endif
